# Использовать постраничный вывод значений в выпадающем списке
DEFAULT_FIELD_IS_PAGINATED = True

# Минимальное кол-во символов, которое необходимо ввести в поле формы для начала
# отправки запроса для получения данных в выпадающем списке
DEFAULT_FIELD_MIN_CHARS = 2

# Задержка в миллисекундах между завершением набора текста в поле формы
# и отправкой запроса на получение данных для выпадающего списка
DEFAULT_FIELD_QUERY_DELAY_MS = 1500
